@@ -167,8 +167,8 @@ class ModelBase(Mapping[str, Any]):
 
     @classmethod
     def _get_field_values_fn(
-        cls, exclude: FieldNamesSet = ()
-    ) -> Callable[["ModelBase"], List[Any]]:
+        cls: Type[T], exclude: FieldNamesSet = ()
+    ) -> Callable[[T], List[Any]]:
         env: Dict[str, Any] = dict()
         func = ["def get_field_values(self): return ["]
         for f in fields(cls):
@@ -413,6 +413,19 @@ class ModelBase(Mapping[str, Any]):
             *cls.upsert_sql(cls.insert_multiple_sql(rows))
         )
 
+    @classmethod
+    def _get_equal_ignoring_fn(
+        cls: Type[T], ignore: FieldNamesSet = ()
+    ) -> Callable[[T, T], bool]:
+        env: Dict[str, Any] = dict()
+        func = ["def equal_ignoring(a, b):"]
+        for f in fields(cls):
+            if f.name not in ignore:
+                func.append(f" if a.{f.name} != b.{f.name}: return False")
+        func += [" return True"]
+        exec("\n".join(func), env)
+        return env["equal_ignoring"]
+
     @classmethod
     async def replace_multiple(
         cls: Type[T],
@@ -422,6 +435,10 @@ class ModelBase(Mapping[str, Any]):
         where: Where,
         ignore: FieldNamesSet = (),
     ) -> Tuple[List[T], List[T], List[T]]:
+        equal_ignoring = cls._cached(
+            ("equal_ignoring", tuple(sorted(ignore))),
+            lambda: cls._get_equal_ignoring_fn(ignore),
+        )
         pending = {row.primary_key(): row for row in map(cls.ensure_model, rows)}
 
         updated = []
@@ -434,7 +451,7 @@ class ModelBase(Mapping[str, Any]):
             if pk not in pending:
                 deleted.append(old)
             else:
-                if not equal_ignoring(old, pending[pk], ignore):
+                if not equal_ignoring(old, pending[pk]):
                     updated.append(pending[pk])
                 del pending[pk]
 
@@ -447,14 +464,62 @@ class ModelBase(Mapping[str, Any]):
 
         return created, updated, deleted
 
+    @classmethod
+    def _get_differences_ignoring_fn(
+        cls: Type[T], ignore: FieldNamesSet = ()
+    ) -> Callable[[T, T], List[str]]:
+        env: Dict[str, Any] = dict()
+        func = [
+            "def differences_ignoring(a, b):",
+            " diffs = []",
+        ]
+        for f in fields(cls):
+            if f.name not in ignore:
+                func.append(
+                    f" if a.{f.name} != b.{f.name}: diffs.append({repr(f.name)})"
+                )
+        func += [" return diffs"]
+        exec("\n".join(func), env)
+        return env["differences_ignoring"]
 
-def equal_ignoring(
-    old: Mapping[str, Any], new: Mapping[str, Any], ignore: FieldNamesSet
-) -> bool:
-    keys = (*old.keys(), *new.keys())
-    for key in keys:
-        if key in ignore:
-            continue
-        if old.get(key, None) != new.get(key, None):
-            return False
-    return True
+    @classmethod
+    async def replace_multiple_reporting_differences(
+        cls: Type[T],
+        connection: Connection,
+        rows: Union[Iterable[T], Iterable[Mapping[str, Any]]],
+        *,
+        where: Where,
+        ignore: FieldNamesSet = (),
+    ) -> Tuple[List[T], List[Tuple[T, T, List[str]]], List[T]]:
+        differences_ignoring = cls._cached(
+            ("differences_ignoring", tuple(sorted(ignore))),
+            lambda: cls._get_differences_ignoring_fn(ignore),
+        )
+
+        pending = {row.primary_key(): row for row in map(cls.ensure_model, rows)}
+
+        updated_triples = []
+        deleted = []
+
+        async for old in cls.select_cursor(
+            connection, where=where, order_by=cls.primary_key_names, for_update=True
+        ):
+            pk = old.primary_key()
+            if pk not in pending:
+                deleted.append(old)
+            else:
+                diffs = differences_ignoring(old, pending[pk])
+                if diffs:
+                    updated_triples.append((old, pending[pk], diffs))
+                del pending[pk]
+
+        created = list(pending.values())
+
+        if created or updated_triples:
+            await cls.upsert_multiple(
+                connection, (*created, *(t[1] for t in updated_triples))
+            )
+        if deleted:
+            await cls.delete_multiple(connection, deleted)
+
+        return created, updated_triples, deleted
